=== modified file 'bzrlib/knit.py'
--- bzrlib/knit.py	2006-03-02 01:26:22 +0000
+++ bzrlib/knit.py	2006-03-02 03:12:34 +0000
@@ -76,7 +76,7 @@
 from bzrlib.trace import mutter
 from bzrlib.osutils import contains_whitespace, contains_linebreaks, \
      sha_strings
-from bzrlib.versionedfile import VersionedFile
+from bzrlib.versionedfile import VersionedFile, InterVersionedFile
 from bzrlib.tsort import topo_sort
 
 
@@ -260,6 +260,9 @@
         self._data = _KnitData(transport, relpath + DATA_SUFFIX,
             mode)
 
+    def create_empty(self, name, transport, mode=None):
+        return KnitVersionedFile(transport, name, 'w', self.factory, delta=self.delta)
+
     def versions(self):
         """See VersionedFile.versions."""
         return self._index.get_versions()
@@ -503,103 +506,6 @@
 
         return self.factory.lower_fulltext(KnitContent(new_lines))
 
-    def join(self, other, pb=None, msg=None, version_ids=None):
-        """See VersionedFile.join."""
-        assert isinstance(other, KnitVersionedFile)
-
-        if version_ids is None:
-            version_ids = other.versions()
-        if not version_ids:
-            return 0
-
-        if pb is None:
-            from bzrlib.progress import DummyProgress
-            pb = DummyProgress()
-
-        version_ids = list(version_ids)
-        if None in version_ids:
-            version_ids.remove(None)
-
-        other_ancestry = set(other.get_ancestry(version_ids))
-        this_versions = set(self._index.get_versions())
-        needed_versions = other_ancestry - this_versions
-        cross_check_versions = other_ancestry.intersection(this_versions)
-        mismatched_versions = set()
-        for version in cross_check_versions:
-            # scan to include needed parents.
-            n1 = set(self.get_parents(version))
-            n2 = set(other.get_parents(version))
-            if n1 != n2:
-                # FIXME TEST this check for cycles being introduced works
-                # the logic is we have a cycle if in our graph we are an
-                # ancestor of any of the n2 revisions.
-                for parent in n2:
-                    if parent in n1:
-                        # safe
-                        continue
-                    else:
-                        parent_ancestors = other.get_ancestry(parent)
-                        if version in parent_ancestors:
-                            raise errors.GraphCycleError([parent, version])
-                # ensure this parent will be available later.
-                new_parents = n2.difference(n1)
-                needed_versions.update(new_parents.difference(this_versions))
-                mismatched_versions.add(version)
-
-        if not needed_versions and not cross_check_versions:
-            return 0
-        full_list = topo_sort(other._index.get_graph())
-
-        version_list = [i for i in full_list if (not self.has_version(i)
-                        and i in needed_versions)]
-
-        records = []
-        for version_id in version_list:
-            data_pos, data_size = other._index.get_position(version_id)
-            records.append((version_id, data_pos, data_size))
-
-        count = 0
-        for version_id, lines, digest \
-                in other._data.read_records_iter(records):
-            options = other._index.get_options(version_id)
-            parents = other._index.get_parents(version_id)
-            
-            for parent in parents:
-                assert self.has_version(parent)
-
-            if self.factory.annotated:
-                # FIXME jrydberg: it should be possible to skip
-                # re-annotating components if we know that we are
-                # going to pull all revisions in the same order.
-                new_version_id = version_id
-                new_version_idx = self._index.num_versions()
-                if 'fulltext' in options:
-                    lines = self._reannotate_fulltext(other, lines,
-                        new_version_id, new_version_idx)
-                elif 'line-delta' in options:
-                    lines = self._reannotate_line_delta(other, lines,
-                        new_version_id, new_version_idx)
-
-            count = count + 1
-            pb.update(self.filename, count, len(version_list))
-
-            pos, size = self._data.add_record(version_id, digest, lines)
-            self._index.add_version(version_id, options, pos, size, parents)
-
-        for version in mismatched_versions:
-            n1 = set(self.get_parents(version))
-            n2 = set(other.get_parents(version))
-            # write a combined record to our history.
-            new_parents = self.get_parents(version) + list(n2.difference(n1))
-            current_values = self._index._cache[version]
-            self._index.add_version(version,
-                                    current_values[1], 
-                                    current_values[2],
-                                    current_values[3],
-                                    new_parents)
-        pb.clear()
-        return count
-
     def walk(self, version_ids):
         """See VersionedFile.walk."""
         # We take the short path here, and extract all relevant texts
@@ -630,7 +536,7 @@
         self._mode = mode
 
     def write_header(self):
-        old_len = self._transport.append(self._filename, self.HEADER)
+        old_len = self._transport.append(self._filename, StringIO(self.HEADER))
         if old_len != 0:
             raise KnitCorrupt(self._filename, 'misaligned after writing header')
 
@@ -745,7 +651,7 @@
                                         size,
                                         ' '.join([str(self.lookup(vid)) for 
                                                   vid in parents]))
-        self._transport.append(self._filename, content)
+        self._transport.append(self._filename, StringIO(content))
 
     def has_version(self, version_id):
         """True if the version is in the index."""
@@ -811,7 +717,7 @@
         data_file.close()
 
         content = sio.getvalue()
-        start_pos = self._transport.append(self._filename, content)
+        start_pos = self._transport.append(self._filename, StringIO(content))
         return start_pos, len(content)
 
     def _parse_record(self, version_id, data):
@@ -889,3 +795,118 @@
             components[record_id] = (content, digest)
         return components
 
+
+class InterKnit(InterVersionedFile):
+    """Optimised code paths for knit to knit operations."""
+    
+    _matching_file_factory = staticmethod(AnnotatedKnitFactory)
+    
+    @staticmethod
+    def is_compatible(source, target):
+        """Be compatible with knits.  """
+        try:
+            return (isinstance(source, KnitVersionedFile) and
+                    isinstance(target, KnitVersionedFile))
+        except AttributeError:
+            return False
+
+    def join(self, pb=None, msg=None, version_ids=None):
+        """See InterVersionedFile.join."""
+        assert isinstance(self.source, KnitVersionedFile)
+        assert isinstance(self.target, KnitVersionedFile)
+
+        if version_ids is None:
+            version_ids = self.source.versions()
+        if not version_ids:
+            return 0
+
+        if pb is None:
+            from bzrlib.progress import DummyProgress
+            pb = DummyProgress()
+
+        version_ids = list(version_ids)
+        if None in version_ids:
+            version_ids.remove(None)
+
+        self.source_ancestry = set(self.source.get_ancestry(version_ids))
+        this_versions = set(self.target._index.get_versions())
+        needed_versions = self.source_ancestry - this_versions
+        cross_check_versions = self.source_ancestry.intersection(this_versions)
+        mismatched_versions = set()
+        for version in cross_check_versions:
+            # scan to include needed parents.
+            n1 = set(self.target.get_parents(version))
+            n2 = set(self.source.get_parents(version))
+            if n1 != n2:
+                # FIXME TEST this check for cycles being introduced works
+                # the logic is we have a cycle if in our graph we are an
+                # ancestor of any of the n2 revisions.
+                for parent in n2:
+                    if parent in n1:
+                        # safe
+                        continue
+                    else:
+                        parent_ancestors = self.source.get_ancestry(parent)
+                        if version in parent_ancestors:
+                            raise errors.GraphCycleError([parent, version])
+                # ensure this parent will be available later.
+                new_parents = n2.difference(n1)
+                needed_versions.update(new_parents.difference(this_versions))
+                mismatched_versions.add(version)
+
+        if not needed_versions and not cross_check_versions:
+            return 0
+        full_list = topo_sort(self.source._index.get_graph())
+
+        version_list = [i for i in full_list if (not self.target.has_version(i)
+                        and i in needed_versions)]
+
+        records = []
+        for version_id in version_list:
+            data_pos, data_size = self.source._index.get_position(version_id)
+            records.append((version_id, data_pos, data_size))
+
+        count = 0
+        for version_id, lines, digest \
+                in self.source._data.read_records_iter(records):
+            options = self.source._index.get_options(version_id)
+            parents = self.source._index.get_parents(version_id)
+            
+            for parent in parents:
+                assert self.target.has_version(parent)
+
+            if self.target.factory.annotated:
+                # FIXME jrydberg: it should be possible to skip
+                # re-annotating components if we know that we are
+                # going to pull all revisions in the same order.
+                new_version_id = version_id
+                new_version_idx = self.target._index.num_versions()
+                if 'fulltext' in options:
+                    lines = self.target._reannotate_fulltext(self.source, lines,
+                        new_version_id, new_version_idx)
+                elif 'line-delta' in options:
+                    lines = self.target._reannotate_line_delta(self.source, lines,
+                        new_version_id, new_version_idx)
+
+            count = count + 1
+            pb.update(self.target.filename, count, len(version_list))
+
+            pos, size = self.target._data.add_record(version_id, digest, lines)
+            self.target._index.add_version(version_id, options, pos, size, parents)
+
+        for version in mismatched_versions:
+            n1 = set(self.target.get_parents(version))
+            n2 = set(self.source.get_parents(version))
+            # write a combined record to our history.
+            new_parents = self.target.get_parents(version) + list(n2.difference(n1))
+            current_values = self.target._index._cache[version]
+            self.target._index.add_version(version,
+                                    current_values[1], 
+                                    current_values[2],
+                                    current_values[3],
+                                    new_parents)
+        pb.clear()
+        return count
+
+
+InterVersionedFile.register_optimiser(InterKnit)

=== modified file 'bzrlib/tests/interversionedfile_implementations/test_join.py'
--- bzrlib/tests/interversionedfile_implementations/test_join.py	2006-03-02 01:26:22 +0000
+++ bzrlib/tests/interversionedfile_implementations/test_join.py	2006-03-02 03:12:34 +0000
@@ -17,8 +17,10 @@
 """Tests for join between versioned files."""
 
 
+import bzrlib.errors as errors
 from bzrlib.tests import TestCaseWithTransport
 from bzrlib.transport import get_transport
+import bzrlib.versionedfile as versionedfile
 
 
 class TestJoin(TestCaseWithTransport):
@@ -30,7 +32,7 @@
         return self.versionedfile_factory(name,
                                           get_transport(self.get_url()))
 
-    def get_target(self, name='source'):
+    def get_target(self, name='target'):
         """"Get an empty versioned file to join into."""
         return self.versionedfile_factory_to(name,
                                              get_transport(self.get_url()))
@@ -47,7 +49,7 @@
         verify_file(f2)
         verify_file(self.get_target())
 
-        self.assertRaises(RevisionNotPresent,
+        self.assertRaises(errors.RevisionNotPresent,
             f2.join, f1, version_ids=['r3'])
 
         #f3 = self.get_file('1')
@@ -58,3 +60,132 @@
         #self.assertTrue(f4.has_version('r0'))
         #self.assertFalse(f4.has_version('r1'))
 
+    def test_gets_expected_inter_worker(self):
+        source = self.get_source()
+        target = self.get_target()
+        inter = versionedfile.InterVersionedFile.get(source, target)
+        self.assertTrue(isinstance(inter, self.interversionedfile_class))
+        
+    def test_join_add_parents(self):
+        """Join inserting new parents into existing versions
+        
+        The new version must have the right parent list and must identify
+        lines originating in another parent.
+        """
+        w1 = self.get_target('w1')
+        w2 = self.get_source('w2')
+        w1.add_lines('v-1', [], ['line 1\n'])
+        w2.add_lines('v-2', [], ['line 2\n'])
+        w1.add_lines('v-3', ['v-1'], ['line 1\n'])
+        w2.add_lines('v-3', ['v-2'], ['line 1\n'])
+        w1.join(w2)
+        self.assertEqual(sorted(w1.versions()),
+                         'v-1 v-2 v-3'.split())
+        self.assertEqualDiff(w1.get_text('v-3'),
+                'line 1\n')
+        self.assertEqual(sorted(w1.get_parents('v-3')),
+                ['v-1', 'v-2'])
+        ann = list(w1.annotate('v-3'))
+        self.assertEqual(len(ann), 1)
+        self.assertEqual(ann[0][0], 'v-1')
+        self.assertEqual(ann[0][1], 'line 1\n')
+        
+    def build_weave1(self):
+        weave1 = self.get_source()
+        self.lines1 = ['hello\n']
+        self.lines3 = ['hello\n', 'cruel\n', 'world\n']
+        weave1.add_lines('v1', [], self.lines1)
+        weave1.add_lines('v2', ['v1'], ['hello\n', 'world\n'])
+        weave1.add_lines('v3', ['v2'], self.lines3)
+        return weave1
+        
+    def test_join_with_empty(self):
+        """Reweave adding empty weave"""
+        wb = self.get_target()
+        w1 = self.build_weave1()
+        w1.join(wb)
+        self.verify_weave1(w1)
+
+    def verify_weave1(self, w1):
+        self.assertEqual(sorted(w1.versions()), ['v1', 'v2', 'v3'])
+        self.assertEqual(w1.get_lines('v1'), ['hello\n'])
+        self.assertEqual([], w1.get_parents('v1'))
+        self.assertEqual(w1.get_lines('v2'), ['hello\n', 'world\n'])
+        self.assertEqual(['v1'], w1.get_parents('v2'))
+        self.assertEqual(w1.get_lines('v3'), ['hello\n', 'cruel\n', 'world\n'])
+        self.assertEqual(['v2'], w1.get_parents('v3'))
+
+    def test_join_with_ghosts_merges_parents(self):
+        """Join combined parent lists"""
+        wa = self.build_weave1()
+        wb = self.get_target()
+        wb.add_lines('x1', [], ['line from x1\n'])
+        wb.add_lines('v1', [], ['hello\n'])
+        wb.add_lines('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
+        wa.join(wb)
+        self.assertEqual(['v1','x1'], wa.get_parents('v2'))
+
+    def test_join_with_ghosts(self):
+        """Join that inserts parents of an existing revision.
+
+        This can happen when merging from another branch who
+        knows about revisions the destination does not.  In 
+        this test the second weave knows of an additional parent of 
+        v2.  Any revisions which are in common still have to have the 
+        same text.
+        """
+        w1 = self.build_weave1()
+        wb = self.get_target()
+        wb.add_lines('x1', [], ['line from x1\n'])
+        wb.add_lines('v1', [], ['hello\n'])
+        wb.add_lines('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
+        w1.join(wb)
+        eq = self.assertEquals
+        eq(sorted(w1.versions()), ['v1', 'v2', 'v3', 'x1',])
+        eq(w1.get_text('x1'), 'line from x1\n')
+        eq(w1.get_lines('v2'), ['hello\n', 'world\n'])
+        eq(w1.get_parents('v2'), ['v1', 'x1'])
+
+    def build_source_weave(self, name, *pattern):
+        w = self.get_source(name)
+        for version, parents in pattern:
+            w.add_lines(version, parents, [])
+        return w
+
+    def build_target_weave(self, name, *pattern):
+        w = self.get_target(name)
+        for version, parents in pattern:
+            w.add_lines(version, parents, [])
+        return w
+
+    def test_join_reorder(self):
+        """Reweave requiring reordering of versions.
+
+        Weaves must be stored such that parents come before children.  When
+        reweaving, we may add new parents to some children, but it is required
+        that there must be *some* valid order that can be found, otherwise the
+        ancestries are contradictory.  (For the specific case of inserting
+        ghost revisions there will be no disagreement, only partial knowledge
+        of the history.)
+
+        Note that the weaves are only partially ordered: when there are two
+        versions where neither is an ancestor of the other the order in which
+        they occur is unconstrained.  When we join those versions into
+        another weave, they may become more constrained and it may be
+        necessary to change their order.
+
+        One simple case of this is 
+
+        w1: (c[], a[], b[a])
+        w2: (b[], c[b], a[])
+        
+        We need to recognize that the final weave must show the ordering
+        a[], b[a], c[b].  The version that must be first in the result is 
+        not first in either of the input weaves.
+        """
+        w1 = self.build_target_weave('1', ('c', []), ('a', []), ('b', ['a']))
+        w2 = self.build_source_weave('2', ('b', []), ('c', ['b']), ('a', []))
+        w1.join(w2)
+        self.assertEqual([], w1.get_parents('a'))
+        self.assertEqual(['a'], w1.get_parents('b'))
+        self.assertEqual(['b'], w1.get_parents('c'))

=== modified file 'bzrlib/tests/test_knit.py'
--- bzrlib/tests/test_knit.py	2006-02-28 08:34:43 +0000
+++ bzrlib/tests/test_knit.py	2006-03-02 03:12:34 +0000
@@ -25,6 +25,7 @@
 from bzrlib.osutils import split_lines
 from bzrlib.tests import TestCaseInTempDir
 from bzrlib.transport.local import LocalTransport
+from bzrlib.transport.memory import MemoryTransport
 from bzrlib.transactions import PassThroughTransaction
 
 
@@ -254,6 +255,16 @@
         self.assertEquals(origins[0], ('text-c', 'z\n')) 
         self.assertEquals(origins[1], ('text-b', 'c\n')) 
 
+    def test_create_empty_annotated(self):
+        k1 = KnitVersionedFile(LocalTransport('.'), 'test1', 'w',
+                               KnitAnnotateFactory())
+        # 0
+        k1.add_lines('text-a', [], ['a\n', 'b\n'])
+        k2 = k1.create_empty('t', MemoryTransport())
+        self.assertTrue(isinstance(k2.factory, KnitAnnotateFactory))
+        self.assertEqual(k1.delta, k2.delta)
+        # the generic test checks for empty content and file class
+
 
 TEXT_1 = """\
 Banana cup cakes:

=== modified file 'bzrlib/tests/test_versionedfile.py'
--- bzrlib/tests/test_versionedfile.py	2006-03-02 01:26:22 +0000
+++ bzrlib/tests/test_versionedfile.py	2006-03-02 03:12:34 +0000
@@ -17,9 +17,6 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-# Remaing to do is to figure out if get_graph should return a simple
-# map, or a graph object of some kind.
-
 
 import bzrlib
 import bzrlib.errors as errors
@@ -33,6 +30,7 @@
 from bzrlib.tests import TestCaseWithTransport
 from bzrlib.trace import mutter
 from bzrlib.transport.local import LocalTransport
+from bzrlib.transport.memory import MemoryTransport
 import bzrlib.versionedfile as versionedfile
 from bzrlib.weave import WeaveFile
 from bzrlib.weavefile import read_weave
@@ -104,6 +102,25 @@
         verify_file(f)
         verify_file(self.reopen_file())
 
+    def test_create_empty(self):
+        f = self.get_file()
+        f.add_lines('0', [], ['a\n'])
+        new_f = f.create_empty('t', MemoryTransport())
+        # smoke test, specific types should check it is honoured correctly for
+        # non type attributes
+        self.assertEqual([], new_f.versions())
+        self.assertTrue(isinstance(new_f, f.__class__))
+
+    def test_get_graph(self):
+        f = self.get_file()
+        f.add_lines('v1', [], ['hello\n'])
+        f.add_lines('v2', ['v1'], ['hello\n', 'world\n'])
+        f.add_lines('v3', ['v2'], ['hello\n', 'cruel\n', 'world\n'])
+        self.assertEqual({'v1': [],
+                          'v2': ['v1'],
+                          'v3': ['v2']},
+                         f.get_graph())
+
     def test_get_parents(self):
         f = self.get_file()
         f.add_lines('r0', [], ['a\n', 'b\n'])
@@ -156,7 +173,6 @@
         self.assertEqual('hello\n', w.get_text('v1'))
         self.assertRaises(errors.WeaveInvalidChecksum, w.get_text, 'v2')
         self.assertRaises(errors.WeaveInvalidChecksum, w.get_lines, 'v2')
-        self.assertRaises(errors.WeaveInvalidChecksum, list, w.get_iter('v2'))
         self.assertRaises(errors.WeaveInvalidChecksum, w.check)
 
         w = self.get_file_corrupted_checksum()
@@ -164,7 +180,6 @@
         self.assertEqual('hello\n', w.get_text('v1'))
         self.assertRaises(errors.WeaveInvalidChecksum, w.get_text, 'v2')
         self.assertRaises(errors.WeaveInvalidChecksum, w.get_lines, 'v2')
-        self.assertRaises(errors.WeaveInvalidChecksum, list, w.get_iter('v2'))
         self.assertRaises(errors.WeaveInvalidChecksum, w.check)
 
     def get_file_corrupted_text(self):
@@ -175,120 +190,6 @@
         """Open the versioned file from disk again."""
         raise NotImplementedError(self.reopen_file)
 
-    def test_join_add_parents(self):
-        """Reweave inserting new parents
-        
-        The new version must have the right parent list and must identify
-        lines originating in another parent.
-        """
-        w1 = self.get_file('w1')
-        w2 = self.get_file('w2')
-        w1.add_lines('v-1', [], ['line 1\n'])
-        w2.add_lines('v-2', [], ['line 2\n'])
-        w1.add_lines('v-3', ['v-1'], ['line 1\n'])
-        w2.add_lines('v-3', ['v-2'], ['line 1\n'])
-        w1.join(w2)
-        self.assertEqual(sorted(w1.names()),
-                         'v-1 v-2 v-3'.split())
-        self.assertEqualDiff(w1.get_text('v-3'),
-                'line 1\n')
-        self.assertEqual(sorted(w1.parent_names('v-3')),
-                ['v-1', 'v-2'])
-        ann = list(w1.annotate('v-3'))
-        self.assertEqual(len(ann), 1)
-        self.assertEqual(ann[0][0], 'v-1')
-        self.assertEqual(ann[0][1], 'line 1\n')
-        
-    def build_weave1(self):
-        weave1 = self.get_file()
-        self.lines1 = ['hello\n']
-        self.lines3 = ['hello\n', 'cruel\n', 'world\n']
-        weave1.add_lines('v1', [], self.lines1)
-        weave1.add_lines('v2', ['v1'], ['hello\n', 'world\n'])
-        weave1.add_lines('v3', ['v2'], self.lines3)
-        return weave1
-        
-    def test_join_with_empty(self):
-        """Reweave adding empty weave"""
-        wb = self.get_file('b')
-        w1 = self.build_weave1()
-        w1.join(wb)
-        self.assertEqual(sorted(w1.iter_names()), ['v1', 'v2', 'v3'])
-        self.assertEqual(w1.get_lines('v1'), ['hello\n'])
-        self.assertEqual([], w1.get_parents('v1'))
-        self.assertEqual(w1.get_lines('v2'), ['hello\n', 'world\n'])
-        self.assertEqual(['v1'], w1.get_parents('v2'))
-        self.assertEqual(w1.get_lines('v3'), ['hello\n', 'cruel\n', 'world\n'])
-        self.assertEqual(['v2'], w1.get_parents('v3'))
-
-    def test_join_with_ghosts_raises_parent_mismatch(self):
-        """Join combined parent lists"""
-        wa = self.build_weave1()
-        wb = self.get_file('b')
-        wb.add_lines('x1', [], ['line from x1\n'])
-        wb.add_lines('v1', [], ['hello\n'])
-        wb.add_lines('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
-        wa.join(wb)
-        self.assertEqual(['v1','x1'], wa.get_parents('v2'))
-
-    def test_join_with_ghosts(self):
-        """Join that inserts parents of an existing revision.
-
-        This can happen when merging from another branch who
-        knows about revisions the destination does not.  In 
-        this test the second weave knows of an additional parent of 
-        v2.  Any revisions which are in common still have to have the 
-        same text.
-        """
-        w1 = self.build_weave1()
-        wb = self.get_file('b')
-        wb.add_lines('x1', [], ['line from x1\n'])
-        wb.add_lines('v1', [], ['hello\n'])
-        wb.add_lines('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
-        w1.join(wb)
-        eq = self.assertEquals
-        eq(sorted(w1.iter_names()), ['v1', 'v2', 'v3', 'x1',])
-        eq(w1.get_text('x1'), 'line from x1\n')
-        eq(w1.get_lines('v2'), ['hello\n', 'world\n'])
-        eq(w1.parent_names('v2'), ['v1', 'x1'])
-
-    def build_empty_weave(self, name, *pattern):
-        w = self.get_file(name)
-        for version, parents in pattern:
-            w.add_lines(version, parents, [])
-        return w
-
-    def test_join_reorder(self):
-        """Reweave requiring reordering of versions.
-
-        Weaves must be stored such that parents come before children.  When
-        reweaving, we may add new parents to some children, but it is required
-        that there must be *some* valid order that can be found, otherwise the
-        ancestries are contradictory.  (For the specific case of inserting
-        ghost revisions there will be no disagreement, only partial knowledge
-        of the history.)
-
-        Note that the weaves are only partially ordered: when there are two
-        versions where neither is an ancestor of the other the order in which
-        they occur is unconstrained.  When we join those versions into
-        another weave, they may become more constrained and it may be
-        necessary to change their order.
-
-        One simple case of this is 
-
-        w1: (c[], a[], b[a])
-        w2: (b[], c[b], a[])
-        
-        We need to recognize that the final weave must show the ordering
-        a[], b[a], c[b].  The version that must be first in the result is 
-        not first in either of the input weaves.
-        """
-        w1 = self.build_empty_weave('1', ('c', []), ('a', []), ('b', ['a']))
-        w2 = self.build_empty_weave('2', ('b', []), ('c', ['b']), ('a', []))
-        w1.join(w2)
-        self.assertEqual([], w1.get_parents('a'))
-        self.assertEqual(['a'], w1.get_parents('b'))
-        self.assertEqual(['b'], w1.get_parents('c'))
 
 class TestWeave(TestCaseWithTransport, VersionedFileTestMixIn):
 
@@ -297,8 +198,8 @@
 
     def get_file_corrupted_text(self):
         w = WeaveFile('foo', LocalTransport('.'))
-        w.add('v1', [], ['hello\n'])
-        w.add('v2', ['v1'], ['hello\n', 'there\n'])
+        w.add_lines('v1', [], ['hello\n'])
+        w.add_lines('v2', ['v1'], ['hello\n', 'there\n'])
         
         # We are going to invasively corrupt the text
         # Make sure the internals of weave are the same

=== modified file 'bzrlib/tree.py'
--- bzrlib/tree.py	2006-01-19 23:14:42 +0000
+++ bzrlib/tree.py	2006-03-02 03:12:34 +0000
@@ -129,19 +129,17 @@
         self._revision_id = revision_id
 
     def get_weave(self, file_id):
-        import bzrlib.transactions as transactions
         return self._weave_store.get_weave(file_id,
                 self._branch.get_transaction())
 
     def get_weave_prelude(self, file_id):
-        import bzrlib.transactions as transactions
         return self._weave_store.get_weave_prelude(file_id,
                 self._branch.get_transaction())
 
     def get_file_lines(self, file_id):
         ie = self._inventory[file_id]
         weave = self.get_weave(file_id)
-        return weave.get(ie.revision)
+        return weave.get_lines(ie.revision)
 
     def get_file_text(self, file_id):
         return ''.join(self.get_file_lines(file_id))

=== modified file 'bzrlib/versionedfile.py'
--- bzrlib/versionedfile.py	2006-03-02 01:26:22 +0000
+++ bzrlib/versionedfile.py	2006-03-02 03:12:34 +0000
@@ -30,6 +30,8 @@
 
 from bzrlib.inter import InterObject
 from bzrlib.symbol_versioning import *
+from bzrlib.transport.memory import MemoryTransport
+from bzrlib.tsort import topo_sort
 
 
 class VersionedFile(object):
@@ -85,6 +87,15 @@
         already present in file history."""
         raise NotImplementedError(self.clone_text)
 
+    def create_empty(self, name, transport, mode=None):
+        """Create a new versioned file of this exact type.
+
+        :param name: the file name
+        :param transport: the transport
+        :param mode: optional file mode.
+        """
+        raise NotImplementedError(self.create_empty)
+
     def get_text(self, version_id):
         """Return version contents as a text string.
 
@@ -112,12 +123,12 @@
             version_ids = [version_ids]
         raise NotImplementedError(self.get_ancestry)
         
-    def get_graph(self, version_id):
-        """Return a graph.
-
-        Must raise RevisionNotPresent if version is not present in
-        file history."""
-        raise NotImplementedError(self.get_graph)
+    def get_graph(self):
+        """Return a graph for the entire versioned file."""
+        result = {}
+        for version in self.versions():
+            result[version] = self.get_parents(version)
+        return result
 
     @deprecated_method(zero_eight)
     def parent_names(self, version):
@@ -155,7 +166,7 @@
 
         Must raise RevisionNotPresent if any of the specified versions
         are not present in the other files history."""
-        raise NotImplementedError(self.join)
+        return InterVersionedFile.get(other, self).join(pb, msg, version_ids)
 
     def walk(self, version_ids=None):
         """Walk the versioned file as a weave-like structure, for
@@ -289,6 +300,31 @@
     _optimisers = set()
     """The available optimised InterVersionedFile types."""
 
+    def join(self, pb=None, msg=None, version_ids=None):
+        """Integrate versions from self.source into self.target.
+
+        If version_ids is None all versions from source should be
+        incorporated into this versioned file.
+
+        Must raise RevisionNotPresent if any of the specified versions
+        are not present in the other files history.
+        """
+        # the default join: 
+        # - make a temporary versioned file of type target
+        # - insert the source content into it one at a time
+        # - join them
+        # Make a new target-format versioned file. 
+        temp_source = self.target.create_empty("temp", MemoryTransport())
+        graph = self.source.get_graph()
+        order = topo_sort(graph.items())
+        for version in order:
+            temp_source.add_lines(version,
+                                  self.source.get_parents(version),
+                                  self.source.get_lines(version))
+        
+        # this should hit the native code path for target
+        return self.target.join(temp_source, pb, msg, version_ids)
+
 
 class InterVersionedFileTestProviderAdapter(object):
     """A tool to generate a suite testing multiple inter versioned-file classes.

=== modified file 'bzrlib/weave.py'
--- bzrlib/weave.py	2006-03-01 12:09:59 +0000
+++ bzrlib/weave.py	2006-03-02 03:12:34 +0000
@@ -84,7 +84,7 @@
 from bzrlib.osutils import sha_strings
 from bzrlib.symbol_versioning import *
 from bzrlib.tsort import topo_sort
-from bzrlib.versionedfile import VersionedFile
+from bzrlib.versionedfile import VersionedFile, InterVersionedFile
 from bzrlib.weavefile import _read_weave_v5, write_weave_v5
 
 
@@ -580,7 +580,10 @@
 
     @deprecated_method(zero_eight)
     def get_iter(self, name_or_index):
-        """Deprecated, please do not use. Lookups are not not needed."""
+        """Deprecated, please do not use. Lookups are not not needed.
+        
+        Please use get_lines now.
+        """
         return self._get_iter(self._maybe_lookup(name_or_index))
 
     @deprecated_method(zero_eight)
@@ -706,27 +709,6 @@
         # no lines outside of insertion blocks, that deletions are
         # properly paired, etc.
 
-
-    def join(self, other, pb=None, msg=None, version_ids=None):
-        import sys, time
-        """Integrate versions from other into this weave.
-
-        The resulting weave contains all the history of both weaves; 
-        any version you could retrieve from either self or other can be 
-        retrieved from self after this call.
-
-        It is illegal for the two weaves to contain different values 
-        or different parents for any version.  See also reweave().
-
-        :param other: The other weave to pull into this one
-        :param pb: An optional progress bar
-        :param msg: An optional message to display for progress
-        """
-        try:
-            self._join(other, pb, msg, version_ids)
-        except errors.WeaveParentMismatch:
-            self._reweave(other, pb, msg)
-
     def _join(self, other, pb, msg, version_ids):
         """Worker routine for join()."""
         if not other.versions():
@@ -857,6 +839,9 @@
         super(WeaveFile, self).add_lines(version_id, parents, lines)
         self._save()
 
+    def create_empty(self, name, transport, mode=None):
+        return WeaveFile(name, transport, mode)
+
     def _save(self):
         """Save the weave."""
         sio = StringIO()
@@ -1177,3 +1162,27 @@
     else:
         sys.exit(main(sys.argv))
 
+
+class InterWeave(InterVersionedFile):
+    """Optimised code paths for weave to weave operations."""
+    
+    _matching_file_factory = staticmethod(WeaveFile)
+    
+    @staticmethod
+    def is_compatible(source, target):
+        """Be compatible with weaves."""
+        try:
+            return (isinstance(source, Weave) and
+                    isinstance(target, Weave))
+        except AttributeError:
+            return False
+
+    def join(self, pb=None, msg=None, version_ids=None):
+        """See InterVersionedFile.join."""
+        try:
+            self.target._join(self.source, pb, msg, version_ids)
+        except errors.WeaveParentMismatch:
+            self.target._reweave(self.source, pb, msg)
+
+
+InterVersionedFile.register_optimiser(InterWeave)

